#ifndef BAYESCC2_H
#define BAYESCC2_H

#include <boost/random.hpp>
#include <boost/random/normal_distribution.hpp>
#include <boost/math/distributions/uniform.hpp>


#include "graph.h"
#include <queue>
#include "evaluation.h"
#include "elapsedtime.h"

using namespace std;
using namespace boost;



class BayesCC2
{
public:

    //
    BayesCC2(Graph * _graph, double _alpha, double _beta, double _teta);
    ~BayesCC2();

    //
    set<VertexID> lookupInitialCandidates();
    bool iterate(set<VertexID> &candidates);
    void labelNode(VertexID vertexID);
    double naiveBayesP(int labelId, VertexID vertexID);
    void lookupNextCandidates(set<VertexID> &candidates, set<VertexID> &newCandidates);
    void updateLabelPriorsP();
    void updateJointProbabilities();
    void eval();
    inline double pcheck(double P);
    int topPrior();
    void labelRemainingNodesUsingPriors();

    //
    vector<double> labelPriorsP;
    vector< vector<double> > labelJointP;

    //
    int iterationCount;
    double K;
    double M;
    double nLabels;
    int nChanges;
    int nLabeled;
    bool changed;
    ElapsedTime elapsedTime;

    //
    Graph *graph;
    double alpha;
    double beta;
    double teta;


};

#endif // BAYESCC2_H
