#! /usr/bin/python
# -*- coding: utf-8 -*-
#####################################################################
# Copyright (C) 2012 Michele Dallachiesa - dallachiesa@disi.unitn.it
#####################################################################

import sys
import os
import math
import re
import random

#####################################################################


#####################################################################

actors = dict() # actorName -> actorId
casts = dict() # actors performing in movie: movieId -> set(actorId1,actorId2,..)
movies = dict() # (movieTitle,movieYear) -> movieId
genres = dict() # genreName -> genreId
movieGenres = dict() # movieId -> set(genreId1,genreId2,..)
actorMovieGenre = dict() # actorId -> genreId
actorMoviesCount = dict() # actorId ->number of movies

movieTitles = dict() # 	movieId -> (title,year)
actorNames = dict() # actorId -> name
genreNames = dict() # genreId -> text

edges = dict() # actorId1 -> set(year1,year2,...).

coActors = dict() # (actorId1,actorId2) -> count





actorId = 0
movieId = 0
genreId = 0

#minYear = 10000
#maxYear = 0

minProbab = 1.0
maxProbab = 0.0
genreNames[0] = "Unknown"


#####################################################################

def addActor(name):
	global actorId
	global actors
	
	try:
		return actors[name]
	except KeyError:
		actorId = actorId+1
		actors[name] = actorId
		actorNames[actorId] = name
		return actorId


def addCastMember(movieId,actorId):
	global casts
	assert actorId > 0
	
	try:
		actorMoviesCount[actorId]+= 1
	except KeyError:
		actorMoviesCount[actorId] = 1		
	
	try:
		casts[movieId].add(actorId)
	except KeyError:
		casts[movieId] = set()		
		casts[movieId].add(actorId)
	

def addGenre(genre):
	global genres
	global genreId
	
	try:
		return genres[genre]
	except KeyError:
		genreId = genreId+1
		genres[genre] = genreId
		genreNames[genreId] = genre
		print "Genre id= " + str(genreId) + " name= " +genre
		return genreId


def addMovie(title, year):
	global movieId
	global movies
		
	try:
		return movies[(title,year)]
	except KeyError:
		movieId = movieId+1
		movies[(title,year)] = movieId
		movieTitles[movieId] = (title,year) #(title,year)
		return movieId
	

def topGenre(actorMovieGenres, actorId):
	
	hist = dict()
	
	for genreId in actorMovieGenres:
		try:			
			hist[genreId]=hist[genreId]+1
		except KeyError:
			hist[genreId]=1
			
	topId=0
	topFreq=0
	
	topPrevId=0
	topPrevFreq=0
	
	count = 0
	for (genreId,genreFreq) in hist.iteritems():
		count += genreFreq
		if genreFreq >= topFreq:
			
			if topPrevId != genreId:
				topPrevId=topId
				topPrevFreq=topFreq
				
			topId = genreId
			topFreq = genreFreq	
			
	if topFreq == topPrevFreq and topId != topPrevId: # if top-1 and top-2 frequent genre is different but same frequency , our confidence is very low.
		topId = 0
		
	if topFreq - topPrevFreq < 5:
		topId = 0
		
	return topId
		
	
	
def loadCast(pathname):
	
	f = open(pathname, "rb")
	fileSize = os.path.getsize(pathname)
	
	sectionState = 0
	actorState = 0
	prev = 0
	
	for line in f:
		
		pos = f.tell()*100 /fileSize
		if pos - prev > 1:
			print "[" + str(pos) + "%]: loading actors.. " + str(len(actorNames))
			prev = pos
									
		line = line.strip(" \t\r\n") #.decode('string_escape')
		
		if sectionState == 0 and line.startswith("----\t\t\t------") == True:
			sectionState = 1 # next line is first actor
			continue
			
		if sectionState == 1 and line.startswith("--------------") == True:
			sectionState = 2 # end of actors after this line
			
		if sectionState != 1: # we are before actor records (sectionState == 0) or after actor records (sectionState == 2): skip.
			continue							

		# if we reach this point, we are reading an actor records or we are going to read the next actor records.

		if line == "":  # we finished to read actor records (next line may contain a) next actor or b) "---------" , i.e. sectionState will become 2).

			actorState = 0 # expect a new actor name on next line.
			
			if len(actorMovies) < minActorMovies: # filter on number of actor movies
				continue
			
			actorId = addActor(actorName)					

			actorMovieGenres = []
			for movieId in actorMovies:
				try:
					(movieTitle,movieYear) = movieTitles[movieId]
#					print "Movie " + str(movieTitle) + " genres: " + str(movieGenres[movieId])
					actorMovieGenres += list(movieGenres[movieId])			
					addCastMember(movieId,actorId)
				except KeyError: # we consider this movie only if it has a genre. we already know that its rating is >minRating.
					pass					
														
			actorMovieGenre[actorId] = topGenre(actorMovieGenres, actorId)
						
		#	print "Top-genre for " + actorName + ": " + genreNames[actorGenreId] + "(" + str(actorGenreId) + ") - " +  str(actorMovieGenres) + " - movieIds: " + str(actorMovies)
			
			
			continue


		if actorState == 0:		
			actorState = 1 # we are now in an actor entry.
			actorMovies = set()
			pos = line.index("\t") # actor names are separated by movie-info with a tab.
			actorName = line[0:pos].replace('"', ' ')
			movieInfo = line[pos:].strip(" \t\r\n")	
			
		else:
			movieInfo = line
			
		(movieTitle,movieYear) = parseMovieInfo(movieInfo)
										
		if (movieTitle,movieYear) == (False,False):
			# skip movie. for example, sometimes year is "(????)". or, year not within specified range.
			continue
			
							
		try:						
			movieId = movies[(movieTitle,movieYear)]
			actorMovies.add(movieId)
		except KeyError: # we do not have its rating
			pass
				
					
	print "Loaded " + str(len(actorNames)) + " actors."	
		
		
		
def parseMovieYear(movieInfo):
	
	tt = re.findall("(\(\d+\))|(\(\d+/)",movieInfo)
	
	if(len(tt) == 0):
		return False
	
	for t in tt: # for all matches
		# t is a pair. t[0] contains matches for (\(\d+\)) and t[1] matches for (\(\d+/).
		# since only one of them exists, we take their union.
		
		t = t[0] + t[1]	
		
		t = re.findall("\d+", t)[0] # we extract the first number we find in the string. parentheses, etc. we know that there is always only one number here. so, we take [0].
		movieYear = int(t)
		
		if movieYear >= yearsInterval[0] and movieYear <= yearsInterval[1]: # sanity checks. we pick the first numerical pattern that looks good.
			break
						
	if movieYear < yearsInterval[0] or movieYear > yearsInterval[1]:
			return False
			
	return movieYear
	
	
										
def parseMovieInfo(movieInfo): # used for actors-list.

	if movieInfo.startswith("\""):
		pos = movieInfo[1:].index("\"")
		movieTitle = movieInfo[1:pos+1]
		movieInfo = movieInfo[pos+2:]
	else:
		movieTitle = False

	movieYear = parseMovieYear(movieInfo)	
	if movieYear == False:
		return (False,False)	
	
	if movieTitle == False:
		pos = movieInfo.index("(" + str(movieYear))
		movieTitle = movieInfo[0:pos]
		
	movieTitle = movieTitle.strip(" \t")
			
	return (movieTitle,movieYear)
	

def parseMovieGenreInfo(movieInfo): # used for genres-list.

	if movieInfo.startswith("\""):
		pos = movieInfo[1:].index("\"")
		movieTitle = movieInfo[1:pos+1]
		movieInfo = movieInfo[pos+2:]
	else:
		movieTitle = False
		
	movieYear = parseMovieYear(movieInfo)	
	if movieYear == False:
		return (False,False,False)	

	if movieTitle == False:
		pos = movieInfo.index("(" + str(movieYear))
		movieTitle = movieInfo[0:pos]
		
	movieTitle = movieTitle.strip(" \t")
	
	pos = movieInfo.rindex("\t")
	
	movieGenre = movieInfo[pos+1:]
	
	return (movieTitle,movieYear,movieGenre)


def	addMovieGenre(movieId,genreId):
	global movieGenres
	
	if movieId not in movieGenres:
		movieGenres[movieId] = set()	
	movieGenres[movieId].add(genreId)



def loadRatings(pathname):
	global movies
	
	print "Loading movie ratings..."
	
	f = open(pathname, "rb")
	sectionState = 0
	
	fileSize = os.path.getsize(pathname)	

	prev = 0

	for line in f:		
		
		pos = f.tell()*100 /fileSize
		if pos - prev > 5:
			print "[" + str(pos) + "%]: loading ratings file.. " + str(len(movies))
			prev = pos
		

		line = line.strip(" \t\r\n") #decode('string_escape')

		if line == "" or line[0] == "N":
			continue

		if sectionState == 0 and line.startswith("MOVIE RATINGS REPORT") == True:
			sectionState = 1
			continue

		if sectionState == 1 and line.startswith("--------------") == True:
			sectionState = 2
			continue

		if sectionState != 1:
			continue			
		
		movieVotes = int(line[10:18])
		movieRate = float(line[18:26])
		movieInfo = line[26:]
		
		
		if movieVotes < minMovieVotes:
			continue
			
		(movieTitle,movieYear) = parseMovieInfo(movieInfo)
		
		if (movieTitle,movieYear) == (False,False):
			continue
		
		movieId = addMovie(movieTitle,movieYear)
		
	print "Loaded " + str(len(movies)) + " movie ratings."	
		
		


def loadGenres(pathname):
	
	
	print "Loading movie genres..."
	
	f = open(pathname, "rb")
	
	sectionState = 0
	
	fileSize = os.path.getsize(pathname)	
	
	prev = 0
	
	for line in f:
		
		pos = f.tell()*100 /fileSize
		if pos - prev > 5:
			print "[" + str(pos) + "%]: loading movie-genres file.. " + str(len(movieGenres))
			prev = pos
			
				
		line = line.strip(" \t\r\n").decode('string_escape')
		
		
		if line == "":
			continue

		if sectionState == 0 and line.startswith("8: THE GENRES LIST") == True:
			sectionState = 1
			
		if sectionState != 1:
			continue
			
		
		movieGenreInfo = parseMovieGenreInfo(line)
				
		if movieGenreInfo == (False,False,False): 
			# we erroneously tried to parse "8: THE GENRES LIST\n=========\n".
			# or, the movie year is not within the time interval as defined in the header global vars.
			continue			
			
		(movieTitle,movieYear,movieGenre) = movieGenreInfo
		genreId = addGenre(movieGenre)
				
						
		try:
			movieId = movies[(movieTitle,movieYear)] # loadRatings adds movies with minimum ratings. If it is not already in, we skip this movie.
			addMovieGenre(movieId,genreId)
									
			if movieId % 10000 == 0:
				print  "[" + str(int(float(f.tell()) / fileSize * 100)) + "%]: Added movie-genre " + str(movieId)
				
		except KeyError:
			pass
				

	print "Loaded " + str(len(movieGenres)) + " movie genres."	
	

	
def updateHistogram(actor, actorGenres):
	global actorLabelHistogram
	global genres

	if actor not in actorGenreHistogram:
		actorGenreHistogram[author] = dict()
		for genre in genres.keys():
			actorGenreHistogram[actor][genre] = 0
		
	for actorGenre in actorGenres:
		actorGenreHistogram[actor][actorGenre] = actorGenreHistogram[actor][actorGenre] +1
	
	
def insertEdge(actorId1, actorId2):
		
	assert actorId1 < actorId2		
		
	try:
		coActors[(actorId1,actorId2)]+= 1
	except KeyError:
		coActors[(actorId1,actorId2)]= 1

		
def buildGraph():
	global movies
	global edges
	
	# we just need to compute edge probabilities. we already have the list of nodes: actors.
	
	# clean up
	movies = dict()
	edges = dict()
	
		
	print "Building graph..."

	totalNumberEdges=0
	for (movieId,actorIds) in casts.iteritems():	
		print "Movie " + str(movieId) + ": " + str(len(actorIds)) + " actors"
		totalNumberEdges += math.pow(len(actorIds),2)-len(actorIds)
		
	print "Total number of edges to produce and process: " + str(totalNumberEdges)
		
	print "Processing " + str(len(casts)) + " casts. Adding edges."
	
#	actorMovieGenre[actorId] # contiene il genreId. non esiste per tutti gli actor.
	
	count = 0
	
	for (movieId,actorIds) in casts.iteritems():	
		count = count +1
						
		if count % 100 == 0:
			print "[" + str(int((float(count) / len(casts)) * 100)) + "%]: Processing movie " + str(count) + " coActors: " + str(len(coActors))
																	
		(movieTitle,movieYear) = movieTitles[movieId]
		for actorId1 in actorIds:
			if actorMovieGenre[actorId1] == 0:
				continue
			for actorId2 in actorIds:
				if actorId1 < actorId2: # every pair is enumerated twice with the double-loop.
					if actorMovieGenre[actorId2] == 0:
						continue
					if actorMoviesCount[actorId1] > actorMoviesCountMin and actorMoviesCount[actorId2] > actorMoviesCountMin:
							insertEdge(actorId1, actorId2)


#	print "Processing " + str(len(casts)) + " casts. Pruning edges."
#						
#	for (movieId,actorIds) in casts.iteritems():	
#		count = count +1
#						
#		if count % 100 == 0:
#			print "[" + str(int((float(count) / len(casts)) * 100)) + "%]: Processing movie " + str(count) + " edges: " + str(len(edges))
#	
#		for actorId1 in actorIds:
#			for actorId2 in actorIds:
#				if actorId1 < actorId2: # every pair is enumerated twice with the double-loop.
#					try:
#						if len(edges[(actorId1, actorId2)]) < minCoActorMovies:
#							del edges[(actorId1, actorId2)]
#					except KeyError:
#						pass
		
	
	# note: the number of existing edges may be lower than the number of insertEdge calls,
	# because two actors may have performed more than once during one year.
		
	print "Loaded " + str(len(coActors)) + " edges."
	

def dumpGraph(pathname):
				
	print "Dumping to file '" + pathname + "' ..."
	f = open(pathname, "wb")

	f.write("*labels*\n")

	for genreName,genreId in genres.iteritems():
		f.write(str(genreId) + ",\"" + genreName + "\"\n")

	f.write("\n*nodes*\n")

	for (actorName,actorId) in actors.iteritems():
		try:
			if actorMoviesCount[actorId] > actorMoviesCountMin:
				f.write( str(actorId) + ",\"" + actorName + "\"," + str(actorMovieGenre[actorId]) + "\n")
		except KeyError:
			pass
				
	
	f.write("\n*edges*" + "\n")

	for (actorId1,actorId2) in coActors:
		f.write(str(actorId1) + "," + str(actorId2) + "," + str(edgeProbability2(actorId1, actorId2)) + "\n")

	f.write("\n*eof*\n")

	f.close()	
	
	print "Stats: minProbab=" + str(minProbab) + " maxProbab=" + str(maxProbab)
	


def edgeProbability2(author1, author2):
	global minProbab
	global maxProbab
		
	c = coActors[(author1, author2)]
	
	assert c > 0

 	# *2 because each publication is counted twice: in coAuthoredCount[author1] and in coAuthoredCount[author2]
	P = float(c*2) / float(actorMoviesCount[author1]+actorMoviesCount[author2])
	
	# P is the ratio of published papers by both authors that the two authors co-authored.
		
	assert (P >= 0 and P <= 1)

	minProbab = min(minProbab, P)
	maxProbab = max(maxProbab, P)
	
	return P



def dumpNet(pathname):
	
	print "Dumping to file '" + pathname + "' ..."
	f = open(pathname, "wb")
	
	f.write("*vertices " + str(len(actors)) + "\n")
	
	for k,v in actors.iteritems(): # k: key (label), v: value (id)
		f.write(str(v) + "\t" + "\"" + k + "\""	+ "\n")
	
	f.write("\n*edges\n")
	
	for (id1,id2) in edges:
		f.write(str(id1) + "\t" + str(id2) + "\n")
		
	f.write("\n")
	
	f.close()
	

def main(argv):

	global yearsInterval
	global minMovieVotes
	global minActorMovies
	global minCoActorMovies
	global actorMoviesCountMin

	#
	yearsInterval = [1980,1990] # extremes included.
	minActorMovies = 1
	minCoActorMovies = 1
	actorMoviesCountMin = 1
	minMovieVotes=10000
	
	graphPathname="imdb.dat"
	actorsPathname= "lists/actors.list"
	genresPathname="lists/genres.list"
	ratingsPathname="lists/ratings.list"				
						
	loadRatings(ratingsPathname)		
	loadGenres(genresPathname)
	loadCast(actorsPathname)
	
	
	buildGraph()
	dumpGraph(graphPathname)	


#	assert minProbab != 1.0
#	assert maxProbab != 0.0
	
	

#####################################################################
if __name__ == "__main__":
    sys.exit(main(sys.argv))


